# coding=UTF-8

import constants
import menu_state
import play_state
import pygame
import random
import utils
import data

class Game(object):
    
    def __init__(self):
        # Preinitialize mixer.
        pygame.mixer.pre_init(frequency=44100)
    
        # Initialize PyGame
        pygame.init()
        
        pygame.mouse.set_visible(False)
        
        # Initialize random seed
        random.seed()
        
        # Create set to keep track of which keyboard keys are down
        self.keys_down = set()
        
        # Create clock
        self.clock = pygame.time.Clock()
        self.elapsed_time = 0
    
        self.prepare_fullscreen_data()

        config = menu_state.Config()
        config.load()

        self.temporary_surface_source = pygame.surface.Surface((constants.WINDOW_WIDTH, constants.WINDOW_HEIGHT))

        if config.fullscreen and self.fullscreen_possible:
            self.initialize_fullscreen_mode()
        else:
            self.initialize_windowed_mode()

        self.font_small = dict()
        self.font_normal = dict()
        self.font_big = dict()
        self.font_score = dict()

        for skin_font in [(constants.SKIN_TRON, "antigravbb_reg.ttf", 18, 32, 64, 32),
                          (constants.SKIN_REALISTIC, "Coda-Regular.ttf", 14, 24, 52, 32)]:
            self.font_small[skin_font[0]] = utils.load_font(skin_font[1], skin_font[2])
            self.font_normal[skin_font[0]] = utils.load_font(skin_font[1], skin_font[3])
            self.font_big[skin_font[0]] = utils.load_font(skin_font[1], skin_font[4])
            self.font_score[skin_font[0]] = utils.load_font(skin_font[1], skin_font[5])
        
        self.sounds = dict()
        
        self.joysticks = []
        self.joystick_check_countdown = 0

        self.state_value = None
        
        self.joystick_button_accept = [0, 7] # A / Start
        self.joystick_button_back = [1, 6] # B / Back

        if constants.DEBUG_SKIP_MENU:
            config.difficulty = constants.DEBUG_SKIP_MENU_DIFFICULTY
            config.duration = constants.DEBUG_SKIP_MENU_DURATION
            config.skin = constants.DEBUG_SKIP_MENU_SKIN
            config.music_active = constants.MUSIC_ON_AT_START
            self.load_sounds(config.skin)
            self.state = play_state.PlayState(self, config)
        else:
            self.state = menu_state.MenuState(self, config)

        pygame.display.set_caption("Hammertennis")
        
        self.screen_shake_distance = 0
        self.screen_shake_delay_between_shakes = constants.SCREEN_SHAKE_DELAY_BETWEEN_SHAKES
        self.screen_shake_next_shake_countdown = 0
        self.screen_shake_countdown = 0
        self.screen_shake_current_distance = data.Point()

    def prepare_fullscreen_data(self):
        info = pygame.display.Info()
    
        self.native_width = info.current_w
        self.native_height = info.current_h
        
        if self.native_width != -1:
            modes = pygame.display.list_modes()
            if len(modes) > 0:
                self.native_width = modes[0][0]
                self.native_height = modes[0][1]
                
        self.fullscreen_possible = self.native_width != -1

    def initialize_windowed_mode(self):
        self.fullscreen = False
        self.display = pygame.display.set_mode((constants.WINDOW_WIDTH, constants.WINDOW_HEIGHT))
        self.temporary_surface_target = self.temporary_surface_source
        self.temporary_surface_target_offset_x = 0
        self.temporary_surface_target_offset_y = 0
        self.scaling = 1

    def initialize_fullscreen_mode(self):
        self.fullscreen = True
        self.display = pygame.display.set_mode((self.native_width, self.native_height), pygame.FULLSCREEN | pygame.HWSURFACE)
        
        scale_x = self.native_width / float(constants.WINDOW_WIDTH)
        scale_y = self.native_height / float(constants.WINDOW_HEIGHT)
        if scale_x > scale_y:
            self.fullscreen_scale_width = int(scale_y * constants.WINDOW_WIDTH)
            self.fullscreen_scale_height = self.native_height
            self.scaling = scale_y
        else:
            self.fullscreen_scale_width = self.native_width
            self.fullscreen_scale_height = int(scale_x * constants.WINDOW_HEIGHT)
            self.scaling = scale_x
        
        self.temporary_surface_target_offset_x = round(self.display.get_size()[0] / 2 - self.fullscreen_scale_width / 2)
        self.temporary_surface_target_offset_y = round(self.display.get_size()[1] / 2 - self.fullscreen_scale_height / 2)
        self.temporary_surface_target = pygame.surface.Surface((self.fullscreen_scale_width, self.fullscreen_scale_height))

    @property
    def state(self):
        return self.state_value

    @state.setter
    def state(self, value):
        self.state_updated = True
        
        if self.state_value != None:
            self.state_value.destroy()
            
        self.state_value = value
        
        self.reset_keys()
        self.state_value.init()

    def run(self):
        # Enter main loop.
        while True:
    
            # Limit frame rate and get the elapsed time
            self.elapsed_time = self.clock.tick(constants.FRAME_RATE) / 1000.0

            self.state_updated = False
                
            if not self.update():
                return
            
            if not self.state_updated:
                self.draw();

    def screen_shake(self, distance, duration):
        self.screen_shake_distance = distance
        self.screen_shake_countdown = duration
        self.screen_shake_next_shake_countdown = 0
    
    def screen_shake_reset(self):
        self.screen_shake_distance = 0
        self.screen_shake_countdown = 0
        self.screen_shake_current_distance.set_to(0, 0)
    
    def update(self):
        self.state.update()

        if self.screen_shake_countdown > 0:
            self.screen_shake_countdown -= self.elapsed_time
            if self.screen_shake_countdown <= 0:
                self.screen_shake_reset()
                
        if self.screen_shake_distance > 0:
            self.screen_shake_next_shake_countdown -= self.elapsed_time
            if self.screen_shake_next_shake_countdown <= 0:
                self.screen_shake_next_shake_countdown = self.screen_shake_delay_between_shakes
                self.screen_shake_current_distance.x = (self.screen_shake_distance - random.random() * self.screen_shake_distance * 2) * self.scaling
                self.screen_shake_current_distance.y = (self.screen_shake_distance - random.random() * self.screen_shake_distance * 2) * self.scaling
        
        # Event handling
        pygame.event.pump()

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                return False
                
            elif event.type == pygame.KEYDOWN:
                # Save that the key is down now
                self.keys_down.add(event.key)
                    
            elif event.type == pygame.KEYUP:
                # Remove previously down key
                self.keys_down.discard(event.key)
                
            self.state.handle_event(event)
            
        return True
                    
    def draw(self):
        if self.fullscreen or (self.screen_shake_distance > 0):
            self.state.draw(self.temporary_surface_source)
            
            if self.scaling != 1:
                pygame.transform.smoothscale(self.temporary_surface_source, (self.fullscreen_scale_width, self.fullscreen_scale_height), self.temporary_surface_target)
                
            self.display.blit(self.temporary_surface_target, [self.temporary_surface_target_offset_x + self.screen_shake_current_distance.x, self.temporary_surface_target_offset_y + self.screen_shake_current_distance.y])
        
        else:
            self.state.draw(self.display)

        # Flip screen buffer to show changes
        pygame.display.flip()
    
    def refresh_joysticks_countdown(self):
        self.joystick_check_countdown -= self.elapsed_time
        if self.joystick_check_countdown <= 0:
            self.joystick_check_countdown = constants.JOYSTICK_CHECK_COUNTDOWN
            self.refresh_joysticks()
            return True
        
        return False

    def refresh_joysticks(self):
        pygame.joystick.quit()
        pygame.joystick.init()
        
        for joystick in self.joysticks:
            joystick.quit()
            
        self.joysticks = []
            
        for i in range(pygame.joystick.get_count()):
            joystick = pygame.joystick.Joystick(i)
            joystick.init()
            self.joysticks.append(joystick)
    
    def reset_keys(self):
        self.keys_down.clear()        
    
    def load_sounds(self, skin):
        self.sounds.clear()
        for sound_name in constants.SOUNDS:
            filename = sound_name + ".ogg"
            sound = utils.load_sound(filename, skin)
            
            if sound == None:
                sound = utils.load_sound(filename, constants.SHARED_SKIN_FOLDER)
            
            if sound != None:
                self.sounds[sound_name] = sound

    def play_sound(self, sound_name):
        if self.sounds.has_key(sound_name):
            self.sounds[sound_name].play()
    
def main():
    game = Game()
    game.run()


# Only run if we are the main program, not an import. This is a Python best
# practice.
if __name__ == "__main__": main()
