import engine
import data
import constants
import pygame
import utils
import play_state
import sys
import ConfigParser

class Config():
    def __init__(self):
        self.difficulty = play_state.PlayState.DIFFICULTY_EASY
        self.duration = play_state.PlayState.DURATION_MEDIUM
        self.skin = constants.SKIN_TRON
        self.music_active = True
        self.graphical_effects = 2
        self.fullscreen = False
        
    def load(self):
        config = ConfigParser.RawConfigParser()
        config.read(constants.CONFIG_FILE)

        section = "general"
        if config.has_section(section):
            if config.has_option(section, "difficulty"): self.difficulty = config.getint(section, "difficulty")
            if config.has_option(section, "duration"): self.duration = config.getint(section, "duration")
            if config.has_option(section, "skin"): self.skin = config.get(section, "skin")
            if config.has_option(section, "music_active"): self.music_active = config.getboolean(section, "music_active")
            if config.has_option(section, "graphical_effects"): self.graphical_effects = config.getint(section, "graphical_effects")
            if config.has_option(section, "fullscreen"): self.fullscreen = config.getboolean(section, "fullscreen")
    
    def save(self):
        config = ConfigParser.RawConfigParser()
        
        section = "general"
        config.add_section(section)
        config.set(section, "difficulty", self.difficulty)
        config.set(section, "duration", self.duration)
        config.set(section, "skin", self.skin)
        config.set(section, "music_active", self.music_active)
        config.set(section, "graphical_effects", self.graphical_effects)
        config.set(section, "fullscreen", self.fullscreen)
        
        with open(constants.CONFIG_FILE, 'wb') as configfile:
                config.write(configfile)


class MenuState(engine.State):

    BACKGROUND_COLOR = (250, 250, 250)
    
    def __init__(self, game, config):
        self.game = game
        self.config = config

    def init(self):
        background_filename = "background.png"
        self.skin_background = dict()
        self.skin_color_normal = dict()
        self.skin_color_hover = dict()

        self.skin_background[constants.SKIN_TRON], _ = utils.load_png(background_filename, constants.SKIN_TRON)
        self.skin_color_normal[constants.SKIN_TRON] = (255, 255, 255)
        self.skin_color_hover[constants.SKIN_TRON] = (107, 143, 212)

        self.skin_background[constants.SKIN_REALISTIC], _ = utils.load_png(background_filename, constants.SKIN_REALISTIC)
        self.skin_color_normal[constants.SKIN_REALISTIC] = (255, 255, 255)
        self.skin_color_hover[constants.SKIN_REALISTIC] = (255, 255, 60)

        self.background = self.skin_background[self.config.skin]
        
        self.keys_left = [ pygame.K_LEFT, pygame.K_a ]
        self.keys_right = [ pygame.K_RIGHT, pygame.K_d ]
        self.keys_up = [ pygame.K_UP, pygame.K_w ]
        self.keys_down = [ pygame.K_DOWN, pygame.K_s ]
        self.keys_action = [ pygame.K_SPACE, pygame.K_RETURN ]
        self.keys_back = [ pygame.K_ESCAPE ]
        
        self.entities = engine.EntityGroup()
        
        self.text_menu = engine.Text("Hammertennis", self.game.font_big[self.config.skin], data.Point(constants.WINDOW_WIDTH/2, 90), self.skin_color_normal[self.config.skin], engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_CENTER)
        self.text_menu.font_dict = self.game.font_big
        self.entities.add(self.text_menu)
        
        #text_x = 50
        #text_y = 400
        #text_offset_y = 5
        self.player_text_lines = []
        
        for i in range(2):
            player_text_line = [
                                engine.Text("Player " + str(i + 1) + ":", self.game.font_small[self.config.skin], data.Point(), self.skin_color_normal[self.config.skin], engine.Text.H_ALIGN_LEFT, engine.Text.V_ALIGN_BOTTOM),
                                engine.Text("", self.game.font_small[self.config.skin], data.Point(), self.skin_color_normal[self.config.skin], engine.Text.H_ALIGN_LEFT, engine.Text.V_ALIGN_BOTTOM)
                                ]
            for text in player_text_line:
                text.font_dict = self.game.font_small
                self.entities.add(text)
                
            self.player_text_lines.append(player_text_line)
        
        self.game.load_sounds(self.config.skin)
        
        self.button_lines = []
        self.hovering_positions = []
        self.new_groups = []
        
        if self.game.fullscreen_possible:
            self.line_fullscreen = self.begin_button_line()
            self.add_button("Windowed", False, self.config.fullscreen)
            self.add_button("Fullscreen", True, self.config.fullscreen)

        self.line_graphical_effects = self.begin_button_line()
        self.add_button("No effects", 0, self.config.graphical_effects)
        self.add_button("Some effects", 1, self.config.graphical_effects)
        self.add_button("Full effects", 2, self.config.graphical_effects)
        
        self.line_music = self.begin_button_line()
        self.add_button("Music on", True, self.config.music_active)
        self.add_button("Music off", False, self.config.music_active)

        self.line_skin = self.begin_button_line()
        self.add_button("Futuristic", constants.SKIN_TRON, self.config.skin)
        self.add_button("Realistic", constants.SKIN_REALISTIC, self.config.skin)

        self.line_difficulty = self.begin_button_line(True)
        self.add_button("Beginner", play_state.PlayState.DIFFICULTY_EASY, self.config.difficulty)
        self.add_button("Professional", play_state.PlayState.DIFFICULTY_HARD, self.config.difficulty)

        self.line_duration = self.begin_button_line()
        self.add_button("Short (" + str(play_state.PlayState.DURATION_SHORT) + ")", play_state.PlayState.DURATION_SHORT, self.config.duration)
        self.add_button("Medium (" + str(play_state.PlayState.DURATION_MEDIUM) + ")", play_state.PlayState.DURATION_MEDIUM, self.config.duration)
        self.add_button("Long (" + str(play_state.PlayState.DURATION_LONG) + ")", play_state.PlayState.DURATION_LONG, self.config.duration)

        line_start = self.begin_button_line()
        self.button_start = self.add_button("Start")
        
        self.begin_button_line(True)
        self.button_exit = self.add_button("Exit")

        self.credit_lines = []
        self.add_credits("Programming", "Tobias Wehrum")
        self.add_credits("Art", "Michael Kessler")
        self.add_credits("Sound", "Moritz Ufer")
        self.add_credits("Music", "Kevin MacLeod")
        self.add_credits("Antigrav Font", "Nate Piekos")
        self.add_credits("Coda Font", "Vernon Adams")

        self.adjust_gui()

        self.hovering_line = line_start
        self.get_hovering_button().hover = True

        self.refresh_joysticks()
        
        self.joystick_locked = set()

    def begin_button_line(self, new_group=False):
        self.button_lines.append([])
        self.hovering_positions.append(0)
        
        line_number = len(self.button_lines) - 1
        
        if new_group:
            self.new_groups.append(line_number)
        
        return line_number
        
    def add_button(self, text, button_value=None, selected_value=None):
        button = SelectableText(text, self.game.font_normal[self.config.skin], data.Point(), self.skin_color_normal[self.config.skin], self.skin_color_hover[self.config.skin], engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_TOP)
        button.value = button_value
        button.selectable = button_value != None
        button.selected = button.selectable and (button_value == selected_value)
        self.entities.add(button)
        
        index = len(self.button_lines)-1
        self.button_lines[index].append(button)
        
        if button.selected:
            self.hovering_positions[index] = len(self.button_lines[index]) - 1
        
        return button

    def add_credits(self, discipline, name):
        discipline_text = engine.Text(discipline + ":", self.game.font_small[self.config.skin], data.Point(), self.skin_color_normal[self.config.skin], engine.Text.H_ALIGN_RIGHT, engine.Text.V_ALIGN_BOTTOM)
        name_text = engine.Text(name, self.game.font_small[self.config.skin], data.Point(), self.skin_color_normal[self.config.skin], engine.Text.H_ALIGN_LEFT, engine.Text.V_ALIGN_BOTTOM)
        
        discipline_text.font_dict = self.game.font_small
        name_text.font_dict = self.game.font_small
        
        self.credit_lines.append((discipline_text, name_text))
        self.entities.add(discipline_text)
        self.entities.add(name_text)

    def adjust_gui(self):
        pos_y = 165
        
        button_distance_x = 30
        button_distance_y = 15
        between_areas_extra_distance = 20
        
        if self.config.skin == constants.SKIN_REALISTIC:
            #button_distance_x = 30
            button_distance_y = 10
            between_areas_extra_distance = 16
            
        
        button_height = self.button_lines[0][0].rect.height
        
        i = 0
        
        for button_line in self.button_lines:
            
            if i in self.new_groups:
                pos_y += between_areas_extra_distance
                
            total_width = -button_distance_x
            biggest_width = 0
            button_count = len(button_line)
            
            for button in button_line:
                button_was_selected = button.selected
                button.selected = False
                biggest_width = max(biggest_width, button.rect.width)
                button.selected = button_was_selected
                
            total_width = (button_count * (biggest_width + button_distance_x) - button_distance_x)
            pos_x = constants.WINDOW_WIDTH / 2 - total_width / 2 + biggest_width / 2
            
            for button in button_line:
                button.set_position(pos_x, pos_y)
                pos_x += biggest_width + button_distance_x
                
            pos_y += button_height + button_distance_y

            i += 1
            
        
        border = 10
        distance_x = 4
        distance_y = 0
        credit_height = self.credit_lines[0][0].rect.height

        biggest_credit_name_width = 0
        
        pos_y = constants.WINDOW_HEIGHT - (len(self.credit_lines) - 1) * (credit_height + distance_y) - border
        
        for credit_line in self.credit_lines:
            biggest_credit_name_width = max(biggest_credit_name_width, credit_line[1].rect.width)
            
        for credit_line in self.credit_lines:
            pos_x = constants.WINDOW_WIDTH - biggest_credit_name_width - border
            credit_line[0].set_position(pos_x - distance_x, pos_y)
            credit_line[1].set_position(pos_x, pos_y)
            pos_y += credit_height + distance_y
            
            
        player_text_height = self.player_text_lines[0][0].rect.height
        
        biggest_player_text_name_width = 0
        
        pos_y = constants.WINDOW_HEIGHT - (len(self.player_text_lines) - 1) * (player_text_height + distance_y) - border
        
        for player_text_line in self.player_text_lines:
            biggest_player_text_name_width = max(biggest_player_text_name_width, player_text_line[0].rect.width)
            
        for player_text_line in self.player_text_lines:
            pos_x = border + biggest_player_text_name_width + distance_x
            player_text_line[0].set_position(border, pos_y) # pos_x - distance_x
            player_text_line[1].set_position(pos_x, pos_y)
            pos_y += player_text_height + distance_y
                
    def update(self):
        self.refresh_joysticks_countdown()

        # Update all entities
        self.entities.update()
    
    def handle_event(self, event):
        if event.type == pygame.KEYDOWN:
            if event.key in self.keys_left:
                self.move_selection(-1, 0)
                
            elif event.key in self.keys_right:
                self.move_selection(1, 0)

            if event.key in self.keys_up:
                self.move_selection(0, -1)
                
            elif event.key in self.keys_down:
                self.move_selection(0, 1)
                
            elif event.key in self.keys_action:
                self.do_action()
                
            elif event.key in self.keys_back:
                self.do_back()

            if constants.DEBUG:
                if event.key == pygame.K_F9:
                    self.game.load_sounds(self.config.skin)
                
        elif event.type == pygame.JOYAXISMOTION:
            index = event.joy + event.axis * 10
            value_abs = abs(event.value)
            value_sign = -1 if event.value < 0 else 1
            if value_abs <= constants.JOYSTICK_DEAD_ZONE:
                if index in self.joystick_locked:
                    self.joystick_locked.remove(index)
                    
            elif value_abs >= constants.JOYSTICK_ACCEPT_ZONE:
                if not index in self.joystick_locked:
                    self.joystick_locked.add(index)
                    if event.axis % 2 == 0:
                        self.move_selection(value_sign, 0)
                    else:
                        self.move_selection(0, value_sign)
        
        elif event.type == pygame.JOYHATMOTION:
            self.move_selection(event.value[0], -event.value[1])
        
        elif event.type == pygame.JOYBUTTONDOWN:
            if event.button in self.game.joystick_button_accept:
                self.do_action()
                
    def draw(self, surface):
        surface.blit(self.background, (0, 0))

        # Draw all entities
        self.entities.draw(surface)

    def move_selection(self, x, y):
        self.get_hovering_button().hover = False

        #if x != 0:        
        self.hovering_line = utils.wrap_around_modulo(self.hovering_line + y, len(self.button_lines))
        
        #if y != 0:
        self.get_hovering_button().selected = False
        self.hovering_positions[self.hovering_line] = utils.wrap_around_modulo(self.hovering_positions[self.hovering_line] + x, len(self.button_lines[self.hovering_line]))

        if x != 0:
            if len(self.button_lines[self.hovering_line]) > 1:
                self.game.play_sound(constants.SOUND_MENU_SELECTION)
            else:
                self.game.play_sound(constants.SOUND_MENU_DENIED)
        else:
            self.game.play_sound(constants.SOUND_MENU_MOVED)
            
        current_button = self.get_hovering_button()
        current_button.hover = True
        if current_button.selectable:
            current_button.selected = True

        self.update_values()

    def update_values(self):
        hovering_value = self.get_hovering_button(self.hovering_line).value
        
        if self.hovering_line == self.line_difficulty:
            self.config.difficulty = hovering_value
            
        elif self.hovering_line == self.line_duration:        
            self.config.duration = hovering_value

        elif self.hovering_line == self.line_skin:
            if self.config.skin != hovering_value:
                self.config.skin = hovering_value
                self.update_skin()
                
        elif self.hovering_line == self.line_music:
            self.config.music_active = hovering_value

        elif self.hovering_line == self.line_graphical_effects:
            self.config.graphical_effects = hovering_value

        elif self.hovering_line == self.line_fullscreen:
            if self.game.fullscreen != hovering_value:
                self.config.fullscreen = hovering_value
                if hovering_value:
                    self.game.initialize_fullscreen_mode()
                else:
                    self.game.initialize_windowed_mode()
                    
        self.config.save()

    def update_skin(self):
        self.background = self.skin_background[self.config.skin]
        
        for e in self.entities:
            if isinstance(e, SelectableText):
                e.color_normal = self.skin_color_normal[self.config.skin]
                e.color_hover = self.skin_color_hover[self.config.skin]
                e.font = self.game.font_normal[self.config.skin]
            elif isinstance(e, engine.Text):
                e.color = self.skin_color_normal[self.config.skin]
                e.font = e.font_dict[self.config.skin]
                
        self.adjust_gui()

        self.game.load_sounds(self.config.skin)

    def do_action(self):
        current_button = self.get_hovering_button()
        
        if current_button == self.button_start:
            self.game.play_sound(constants.SOUND_MENU_ACTION)
            self.start_game()
            
        elif current_button == self.button_exit:
            self.exit_game()
            
        else:
            self.game.play_sound(constants.SOUND_MENU_DENIED)
    
    def do_back(self):
        self.exit_game()
    
    def start_game(self):
        self.refresh_joysticks()
        self.game.state = play_state.PlayState(self.game, self.config)
    
    def exit_game(self):
        sys.exit()

    def get_hovering_button(self, line=None):
        if line == None:
            line = self.hovering_line
            
        return self.button_lines[line][self.hovering_positions[line]]
    
    def refresh_joysticks_countdown(self):
        if self.game.refresh_joysticks_countdown():
            self.refresh_joysticks_text()
            
    def refresh_joysticks(self):
        self.game.refresh_joysticks()
        self.refresh_joysticks_text()
        
    def refresh_joysticks_text(self):
        text = ["No gamepad found (Keyboard: WASD)", "No gamepad found (Keyboard: Arrows)"]
        
        for joystick in self.game.joysticks:
            joystick_id = joystick.get_id()
            if joystick_id < len(self.player_text_lines):
                text[joystick_id] = joystick.get_name()

        for i in range(len(self.player_text_lines)):
            self.player_text_lines[i][1].text = text[i]
            
    
class SelectableText(engine.Text):
    
    def __init__(self, text, font, position, color_normal, color_hover, h_align=engine.Text.H_ALIGN_LEFT, v_align=engine.Text.V_ALIGN_TOP):
        self.color_normal_value = color_normal
        self.color_hover_value = color_hover
        
        self.hover_value = False
        self.selected_value = False

        engine.Text.__init__(self, text, font, position, color_normal, h_align, v_align)
        
    @property
    def color_normal(self):
        return self.color_normal_value

    @color_normal.setter
    def color_normal(self, value):
        self.color_normal_value = value
        self.update_hover_color()
        
    @property
    def color_hover(self):
        return self.color_hover_value

    @color_hover.setter
    def color_hover(self, value):
        self.color_hover_value = value
        self.update_hover_color()
        
    @property
    def hover(self):
        return self.hover_value

    @hover.setter
    def hover(self, value):
        self.hover_value = value
        self.update_hover_color()
    
    def update_hover_color(self):
        if self.hover_value:
            self.color = self.color_hover
        else:
            self.color = self.color_normal

    @property
    def selected(self):
        return self.selected_value

    @selected.setter
    def selected(self, value):
        self.selected_value = value
        self.render_text(self.text_value)

    def render_text(self, text):
        if self.selected:
            text = "[" + text + "]"
        
        engine.Text.render_text(self, text)