import data
import random
import math
import utils


class Particle():
    def __init__(self, x, y, velocity_x, velocity_y, acceleration_x, acceleration_y, max_life_time, fade_time):
        self.x = x
        self.y = y
        self.velocity_x = velocity_x
        self.velocity_y = velocity_y
        #self.acceleration_x = acceleration_x
        #self.acceleration_y = acceleration_y
        self.max_life_time = max_life_time
        self.fade_start_at = max_life_time - fade_time
        self.fade_time = fade_time
        self.time = 0

class ParticleSpawnerNone():
    def __init__(self):
        pass
    
    def spawn(self, emitter, x, y):
        pass

class ParticleSpawnerAllDirections():
    def __init__(self, min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count):
        self.min_speed = min_speed
        self.max_speed = max_speed
        self.acceleration_x = acceleration_x
        self.acceleration_y = acceleration_y
        self.max_life_time = max_life_time
        self.fade_time = fade_time
        self.count = count
        
    def spawn(self, emitter, x, y):
        for _ in range(0, self.count):
            speed = self.min_speed + random.random() * (self.max_speed - self.min_speed)
            angle = random.random() * math.pi * 2
            velocity_x = math.cos(angle) * speed
            velocity_y = math.sin(angle) * speed
            emitter.add_particle(Particle(x, y, velocity_x, velocity_y, self.acceleration_x, self.acceleration_y, self.max_life_time, self.fade_time))


class ParticleSpawnerDirectional():
    def __init__(self, min_angle, max_angle, min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count):
        self.min_angle = min_angle
        self.max_angle = max_angle
        self.min_speed = min_speed
        self.max_speed = max_speed
        self.acceleration_x = acceleration_x
        self.acceleration_y = acceleration_y
        self.max_life_time = max_life_time
        self.fade_time = fade_time
        self.count = count
        
        while self.min_angle < 0:
            self.min_angle += math.pi * 2
            self.max_angle += math.pi * 2
        
    def spawn(self, emitter, x, y):
        for _ in range(0, self.count):
            speed = self.min_speed + random.random() * (self.max_speed - self.min_speed)
            angle = self.min_angle + random.random() * (self.max_angle - self.min_angle)
            velocity_x = math.cos(angle) * speed
            velocity_y = math.sin(angle) * speed
            emitter.add_particle(Particle(x, y, velocity_x, velocity_y, self.acceleration_x, self.acceleration_y, self.max_life_time, self.fade_time))


class ParticleSpawnerTrail():
    def __init__(self, max_life_time, fade_time, step_size, start_x, start_y):
        self.max_life_time = max_life_time
        self.fade_time = fade_time
        self.step_size = step_size
        self.previous_x = start_x
        self.previous_y = start_y
        self.previous_vector = data.Point()
        
    def spawn(self, emitter, x, y):
        if self.step_size > 0:
            self.previous_vector.x = x - self.previous_x
            self.previous_vector.y = y - self.previous_y
            
            length = self.previous_vector.length()
            if length > self.step_size * 1.5:
                self.previous_vector.x /= length
                self.previous_vector.y /= length
                
                for distance in utils.frange(self.step_size, length, self.step_size):
                    emitter.add_particle(Particle(self.previous_x + self.previous_vector.x * distance, self.previous_y + self.previous_vector.y * distance, 0, 0, 0, 0, self.max_life_time, self.fade_time))
                
        emitter.add_particle(Particle(x, y, 0, 0, 0, 0, self.max_life_time, self.fade_time))

        self.previous_x = x
        self.previous_y = y


class ParticleEmitter():
    def __init__(self, game, image, active):
        self.game = game
        self.position = data.Point()
        self.image = image
        self.rect = image.get_rect()
        self.origin = data.Point(self.rect.width / 2, self.rect.height / 2) 
        self.active = active
        
        self.particles = []
    
    def add_particle(self, particle):
        if self.active:
            self.particles.append(particle)
    
    def update(self):
        elapsed_time = self.game.elapsed_time
        
        remove_particles = None
        
        for particle in self.particles:
            particle.x += particle.velocity_x * elapsed_time
            particle.y += particle.velocity_y * elapsed_time
            
            #particle.velocity_x += particle.acceleration_x * elapsed_time
            #particle.velocity_y += particle.acceleration_y * elapsed_time
            
            #if particle.x < 0:
            #    particle.x += constants.WINDOW_WIDTH

            #if particle.x >= constants.WINDOW_WIDTH:
            #    particle.x -= constants.WINDOW_WIDTH

            #if particle.y < 0:
            #    particle.y += constants.WINDOW_HEIGHT

            #if particle.y >= constants.WINDOW_HEIGHT:
            #    particle.y -= constants.WINDOW_HEIGHT
                
            if particle.max_life_time > 0:
                particle.time = particle.time + elapsed_time
                
                if particle.time > particle.max_life_time:
                    if remove_particles == None:
                        remove_particles = [particle]
                    else:
                        remove_particles.append(particle)
                        
        if remove_particles != None:
            for particle in remove_particles:
                self.particles.remove(particle)
                

    def draw(self, surface):
        for particle in self.particles:
            self.rect.left = particle.x - self.origin.x 
            self.rect.top = particle.y - self.origin.y
            if (particle.max_life_time > 0) and (particle.fade_time > 0) and (particle.time > particle.fade_start_at):
                self.image.set_alpha(255 - ((particle.time - particle.fade_start_at) / particle.fade_time) * 255)
            else:
                self.image.set_alpha(255)
            surface.blit(self.image, self.rect)

    def handle_event(self, event):
        return