import constants
import utils
import pygame
import random
import data
import Box2D
import debug
import engine
import math
import menu_state
import particles


class PlayState(engine.State, Box2D.b2ContactListener):
    
    DIFFICULTY_EASY = 0
    DIFFICULTY_HARD = 1
    
    DURATION_SHORT = 10
    DURATION_MEDIUM = 20
    DURATION_LONG = 30
    
    def __init__(self, game, config):
        Box2D.b2ContactListener.__init__(self)
        
        self.game = game
        self.config = config
        
    def init(self):
        # Initialize physics
        self.world = Box2D.b2World(gravity=(0, constants.GRAVITY),
                                   doSleep=False,
                                   contactListener=self)
        
        if constants.DEBUG_BOX2D_OUTPUT:
            self.debugDraw = debug.Box2dDebugDraw(surface=self.game.display,
                                                  zoom=constants.WORLD_TO_SCREEN,
                                                  center=(constants.WINDOW_WIDTH/2, constants.WINDOW_HEIGHT/2),
                                                  offset=Box2D.b2Vec2(0, 0),
                                                  screenSize=(constants.WINDOW_WIDTH, constants.WINDOW_HEIGHT))
            self.world.renderer = self.debugDraw
            self.debugDraw.flags=dict(
                    drawShapes=True,
                    drawJoints=True,
                    drawAABBs =True,
                    drawPairs =True,
                    drawCOMs  =True)

        border_size = 100
        border_category = constants.FILTER_BORDER
        border_mask = constants.FILTER_BALL | constants.FILTER_PLAYER | constants.FILTER_HAMMER

        utils.create_static_rectangle_body(self.world, 0, -border_size, constants.WINDOW_WIDTH, border_size, border_category, border_mask)              # top
        utils.create_static_rectangle_body(self.world, 0, constants.WINDOW_HEIGHT, constants.WINDOW_WIDTH, border_size, border_category, border_mask)   # bottom

        # Load the background
        self.background, _ = utils.load_png("background.png", self.config.skin)
    
        # Load music, set the volume and and play
        self.music = utils.load_music("incompetech - Padanaya Blokov.ogg")
        if self.config.music_active:
            self.music.set_volume(constants.MUSIC_VOLUME_FULL)
        else:
            self.music.set_volume(0)
        self.music.play(-1) # Loop indefinitely

        # Initialize particle emitters
        self.emitters_background = []
        self.emitters_foreground = []

        particle_image_red, _ = utils.load_png("particle_red.png", self.config.skin)
        particle_image_red.set_colorkey((0, 0, 0))
        
        particle_image_blue, _ = utils.load_png("particle_blue.png", self.config.skin)
        particle_image_blue.set_colorkey((0, 0, 0))
        
        particle_image_white, _ = utils.load_png("particle_white.png", self.config.skin)
        particle_image_white.set_colorkey((0, 0, 0))
        
        emitter_red = particles.ParticleEmitter(self.game, particle_image_red, self.config.graphical_effects >= 1)
        self.emitters_foreground.append(emitter_red)

        emitter_blue = particles.ParticleEmitter(self.game, particle_image_blue, self.config.graphical_effects >= 1)
        self.emitters_foreground.append(emitter_blue)
        
        emitter_white = particles.ParticleEmitter(self.game, particle_image_white, self.config.graphical_effects >= 1)
        self.emitters_foreground.append(emitter_white)
        
        if self.config.graphical_effects >= 2:
            particle_image_ball_trail_red, _ = utils.load_png("trail_particle_red.png", self.config.skin)
            particle_image_ball_trail_red.set_colorkey((0, 0, 0))
            
            particle_image_ball_trail_blue, _ = utils.load_png("trail_particle_blue.png", self.config.skin)
            particle_image_ball_trail_blue.set_colorkey((0, 0, 0))
            
            particle_image_ball_trail_neutral, _ = utils.load_png("trail_particle_white.png", self.config.skin)
            particle_image_ball_trail_neutral.set_colorkey((0, 0, 0))
            
            emitter_ball_trail_red = particles.ParticleEmitter(self.game, particle_image_ball_trail_red, True)
            self.emitters_background.append(emitter_ball_trail_red)
    
            emitter_ball_trail_blue = particles.ParticleEmitter(self.game, particle_image_ball_trail_blue, True)
            self.emitters_background.append(emitter_ball_trail_blue)
            
            emitter_ball_trail_neutral = particles.ParticleEmitter(self.game, particle_image_ball_trail_neutral, True)
            self.emitters_background.append(emitter_ball_trail_neutral)

        else:
            emitter_ball_trail_red = None
            emitter_ball_trail_blue = None
            emitter_ball_trail_neutral = None
            
        self.ball_default_emitter = emitter_white
        self.ball_trail_default_emitter = emitter_ball_trail_neutral


        if self.config.graphical_effects >= 1:
            
            min_speed, max_speed, max_life_time, fade_time, count = (30, 60, 0.6, 0.2, 10)
            
            right_direction = 0
            up_direction = math.pi * 0.5
            left_direction = math.pi
            down_direction = math.pi * 1.5
            
            wall_angle_range = math.pi / 2
            goal_angle_range = math.pi / 10
            
            min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count = (30, 60, 0, 0, 0.6, 0.2, 10)
            self.particle_spawner_collision_ball_border_down = particles.ParticleSpawnerDirectional(down_direction - wall_angle_range / 2, down_direction + wall_angle_range / 2,
                                                                                                    min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count)
            self.particle_spawner_collision_ball_border_up = particles.ParticleSpawnerDirectional(up_direction - wall_angle_range / 2, up_direction + wall_angle_range / 2,
                                                                                                  min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count)
            
            min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count = (30, 60, 0, 0, 0.6, 0.2, 10)
            self.particle_spawner_collision_ball_hammer = particles.ParticleSpawnerAllDirections(min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count)
            
            min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count = (30, 60, 0, 0, 0.6, 0.2, 10)
            self.particle_spawner_collision_ball_player = particles.ParticleSpawnerAllDirections(min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count)
            
            min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count = (30, 60, 0, 0, 0.6, 0.2, 10)
            self.particle_spawner_collision_hammer_hammer = particles.ParticleSpawnerAllDirections(min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count)
            
            min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count = (75, 150, 0, 0, 1, 0.2, 10)
            #min_speed, max_speed, acceleration_x, acceleration_y, max_life_time, fade_time, count = (200, 350, -250, 0, 1, 0.2, 10)
            self.particle_spawner_goal_left = particles.ParticleSpawnerDirectional(right_direction - goal_angle_range / 2, right_direction + goal_angle_range / 2, min_speed, max_speed,
                                                                                   acceleration_x, acceleration_y, max_life_time,fade_time, count)
            self.particle_spawner_goal_right = particles.ParticleSpawnerDirectional(left_direction - goal_angle_range / 2, left_direction + goal_angle_range / 2, min_speed, max_speed,
                                                                                    -acceleration_x, acceleration_y, max_life_time,fade_time, count)
        else:
            empty_spawner = particles.ParticleSpawnerNone()
            self.particle_spawner_collision_ball_border_down = empty_spawner
            self.particle_spawner_collision_ball_border_up = empty_spawner
            self.particle_spawner_collision_ball_hammer = empty_spawner
            self.particle_spawner_collision_ball_player = empty_spawner
            self.particle_spawner_collision_hammer_hammer = empty_spawner
            self.particle_spawner_goal_left = empty_spawner
            self.particle_spawner_goal_right = empty_spawner
            

        # Initialize entity arrays/groups
        self.players = []
        self.entities = engine.EntityGroup()
        
        player_distance = 140
        hammer_start_offset_x = 80
        
        hammer_max_distance_world = constants.PLAYER_HAMMER_MAX_DISTANCE_WORLD
        if self.config.difficulty == PlayState.DIFFICULTY_HARD:
            hammer_max_distance_world = 0

        player_color_red = (255, 47, 47)
        player_color_blue = (0, 110, 208)
        player_hammer_rope_length_subtraction = 26
        rotate_player = False
        if self.config.skin == constants.SKIN_REALISTIC:
            player_color_red = (181, 64, 64)
            player_color_blue = (55, 139, 139)
            player_hammer_rope_length_subtraction = 0
            rotate_player = True        

        # Initialize left player
        self.player_left = Player(self, "Red Player", "red", player_color_red, constants.FILTER_PLAYER_1, data.Point(player_distance, constants.WINDOW_HEIGHT / 2),
                                  rotate_player, data.Point(-hammer_start_offset_x, 0), player_hammer_rope_length_subtraction, hammer_max_distance_world, emitter_red, emitter_ball_trail_red,
                                  pygame.K_a, pygame.K_d, pygame.K_w, pygame.K_s, 0)
        self.players.append(self.player_left)
        self.entities.add(self.player_left)
        #left_fixture.userData = (constants.USER_DATA_GOAL, player_left)

        # Initialize right player
        self.player_right = Player(self, "Blue Player", "blue", player_color_blue, constants.FILTER_PLAYER_2, data.Point(constants.WINDOW_WIDTH - player_distance, constants.WINDOW_HEIGHT / 2),
                                   rotate_player, data.Point(hammer_start_offset_x, 0), player_hammer_rope_length_subtraction, hammer_max_distance_world, emitter_blue, emitter_ball_trail_blue,
                                   pygame.K_LEFT, pygame.K_RIGHT, pygame.K_UP, pygame.K_DOWN, 1)
        self.players.append(self.player_right)
        self.entities.add(self.player_right)
        #right_fixture.userData = (constants.USER_DATA_GOAL, player_right)
        
        utils.create_static_rectangle_body(self.world, -border_size, 0, border_size, constants.WINDOW_HEIGHT, border_category, border_mask, (constants.USER_DATA_GOAL, self.player_right))          # left
        utils.create_static_rectangle_body(self.world, constants.WINDOW_WIDTH, 0, border_size, constants.WINDOW_HEIGHT, border_category, border_mask, (constants.USER_DATA_GOAL, self.player_left))# right
        
        # Create ball
        self.ball = None
        #self.throw_ball()
        self.respawn_ball_countdown = constants.BALL_THROW_IN_COUNTDOWN_START
        
        score_position_y = 0
        score_additional_y = 5
        score_distance = 10
        additional_score_distance_left = 2
        if self.config.skin == constants.SKIN_REALISTIC:
            score_position_y = 4
            score_additional_y = -1
            additional_score_distance_left = 0
        
        self.score_text_left = engine.Text("", self.game.font_score[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2 - score_distance, score_position_y + score_additional_y),
                                           player_color_red, engine.Text.H_ALIGN_RIGHT, engine.Text.V_ALIGN_TOP)
        self.score_text_center = engine.Text(":", self.game.font_score[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2, score_position_y), (0, 0, 0), engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_TOP)
        self.score_text_right = engine.Text("", self.game.font_score[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2 + score_distance + additional_score_distance_left, score_position_y + score_additional_y),
                                            player_color_blue, engine.Text.H_ALIGN_LEFT, engine.Text.V_ALIGN_TOP)
        self.entities.add(self.score_text_left)
        self.entities.add(self.score_text_center)
        self.entities.add(self.score_text_right)
        
        self.update_score()
        
        self.ball_remove = False
        
        self.overlay = pygame.Surface((constants.WINDOW_WIDTH, constants.WINDOW_HEIGHT))
        self.overlay.set_alpha(100)
        self.overlay.fill((255, 255, 255))
        
        self.overlay_entities = engine.EntityGroup()
        
        self.fps_text = None
        if constants.DEBUG:
            self.fps_text = engine.Text("", self.game.font_small[self.config.skin], data.Point(35, 5), (255, 0, 0), engine.Text.H_ALIGN_LEFT, engine.Text.V_ALIGN_TOP)
            self.overlay_entities.add(self.fps_text)
        
        self.game_over = False
        self.paused = False

        self.pause_entities = []
        
        color = (0, 0, 0)

        pause_text = engine.Text("Paused", self.game.font_big[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2, constants.WINDOW_HEIGHT / 2), color, engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_CENTER)
        self.pause_entities.append(pause_text)
        
        restart_string = "Press P/Escape "
        back_string = "Press Return "
        
        if len(self.game.joysticks) > 0:
            restart_string += "or A/Start "
            back_string += "or B/Back "
        
        restart_string += "to resume."
        back_string += "to go back to the menu."
        
        offset = 10
        y = constants.WINDOW_HEIGHT - offset
        
        back_text = engine.Text(back_string, self.game.font_small[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2, y), color, engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_BOTTOM)
        self.pause_entities.append(back_text)
        
        y = y - back_text.rect.height - offset
        restart_text = engine.Text(restart_string, self.game.font_small[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2, y), color, engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_BOTTOM)
        self.pause_entities.append(restart_text)
        
        self.game.play_sound(constants.SOUND_GAME_STARTS)

        self.hammer_hit_contact = False
        
    def destroy(self):
        self.music.stop()

    def update(self):
        
        if self.fps_text != None:
            self.fps_text.text = "FPS: " + str(int(1/self.game.elapsed_time))
        
        self.overlay_entities.update()

        if self.game_over or self.paused:
            return
        
        #self.world.Step(1.0 / constants.FRAME_RATE * constants.PHYSICS_SPEED, 10, 10)
        self.world.Step(self.game.elapsed_time * constants.PHYSICS_SPEED, 10, 10)
        
        if self.game_over:
            return
        
        if self.ball_remove:
            self.ball_remove = False
            self.entities.remove(self.ball)
            self.world.DestroyBody(self.ball.body)
            self.ball = None
            self.respawn_ball_countdown = constants.BALL_THROW_IN_COUNTDOWN
            
        if self.ball == None:
            self.respawn_ball_countdown -= self.game.elapsed_time
            if self.respawn_ball_countdown <= 0:
                self.throw_ball()
        
        # Update all entities
        self.entities.update()
        
        for emitter in self.emitters_background:
            emitter.update()
        
        for emitter in self.emitters_foreground:
            emitter.update()
        
    def draw(self, surface):
        # Blit the background image to the display.
        surface.blit(self.background, (0, 0))

        for emitter in self.emitters_background:
            emitter.draw(surface)
            
        if constants.DEBUG_BOX2D_OUTPUT:
            self.world.DrawDebugData()
            
        # Draw all entities
        self.entities.draw(surface)

        for emitter in self.emitters_foreground:
            emitter.draw(surface)
        
        if self.paused or self.game_over:
            surface.blit(self.overlay, (0, 0))
        
        self.overlay_entities.draw(surface)
    
    def handle_event(self, event):
        if event.type == pygame.KEYDOWN:
            # Handle music toggle key
            if event.key == pygame.K_m:
                if self.music.get_volume() == 0:
                    self.music.set_volume(constants.MUSIC_VOLUME_FULL)
                else:
                    self.music.set_volume(0)
            
            if self.game_over:
                if event.key == pygame.K_SPACE:
                    self.restart()
                elif event.key == pygame.K_ESCAPE:
                    self.back_to_menu()
                    
            elif self.paused:
                if (event.key == pygame.K_ESCAPE) or (event.key == pygame.K_p) or (event.key == pygame.K_SPACE):
                    self.resume()
                elif event.key == pygame.K_RETURN:
                    self.back_to_menu()
                    
            elif (event.key == pygame.K_p) or (event.key == pygame.K_ESCAPE):
                self.pause()
                
            if constants.DEBUG:
                if event.key == pygame.K_F9:
                    self.game.load_sounds(self.config.skin)
        
        elif event.type == pygame.JOYBUTTONDOWN:
            if self.game_over:
                if event.button in self.game.joystick_button_accept:
                    self.restart()
                elif event.button in self.game.joystick_button_back:
                    self.back_to_menu()
                    
            elif self.paused:
                if event.button in self.game.joystick_button_accept:
                    self.resume()
                elif event.button in self.game.joystick_button_back:
                    self.back_to_menu()
                    
            elif (event.button in self.game.joystick_button_accept):
                self.pause()
                
        
        # Inform the entities of the event
        for entity in self.entities:
            entity.handle_event(event)
            
    def BeginContact(self, contact):
        entities = []
        
        players = 0
        goals = 0
        hammers = 0
        balls = 0
        walls = 0
        
        ball = None
        
        if contact.fixtureA.userData == None:
            walls += 1
        else:
            entities.append(contact.fixtureA.userData)
            
        if contact.fixtureB.userData == None:
            walls += 1
        else:
            entities.append(contact.fixtureB.userData)
        
        player = None
        hammer = None
        
        for (t, e) in entities:
            if isinstance(e, Player):
                if t == constants.USER_DATA_ENTITY:
                    players += 1
                    player = e
                else:
                    goals += 1
            elif isinstance(e, Hammer):
                hammers += 1
                hammer = e
            elif isinstance(e, Ball):
                balls += 1
                ball = e
        
        sound = None
        
        if balls == 1:
            if hammers == 1:
                sound = constants.SOUND_COLLISION_BALL_HAMMER
                self.screen_shake(constants.SCREEN_SHAKE_BALL_HAMMER_DISTANCE, constants.SCREEN_SHAKE_BALL_HAMMER_DURATION)
                
                self.spawn_particles_at_contact_position(contact, self.particle_spawner_collision_ball_hammer, hammer.player.emitter)
                
            elif (walls == 1) and not ball.freshly_thrown_in:
                sound = constants.SOUND_COLLISION_BALL_BORDER
                self.screen_shake(constants.SCREEN_SHAKE_BALL_BORDER_DISTANCE, constants.SCREEN_SHAKE_BALL_BORDER_DURATION)
                
                worldManifold = contact.worldManifold
                point = worldManifold.points[0]
                point = utils.world_to_screen(point[0], point[1])
                
                spawner = self.particle_spawner_collision_ball_border_up
                if point[1] > constants.WINDOW_WIDTH / 2:
                    spawner = self.particle_spawner_collision_ball_border_down
                
                self.spawn_particles(point[0], point[1], spawner, self.ball.get_current_emitter())
                
            elif players == 1:
                sound = constants.SOUND_COLLISION_BALL_PLAYER
                
                self.spawn_particles_at_contact_position(contact, self.particle_spawner_collision_ball_player, player.emitter)
                #self.screen_shake(constants.SCREEN_SHAKE_BALL_PLAYER_DISTANCE, constants.SCREEN_SHAKE_BALL_PLAYER_DURATION)
                
        elif hammers == 2:
            sound = constants.SOUND_COLLISION_HAMMER_HAMMER
            self.screen_shake(constants.SCREEN_SHAKE_HAMMER_HAMMER_DISTANCE, constants.SCREEN_SHAKE_HAMMER_HAMMER_DURATION)
            
            if constants.HAMMER_HAMMER_COLLISON_PARTICLE_OFFSET_DISTANCE == 0:
                _, entity_a = contact.fixtureA.userData
                _, entity_b = contact.fixtureB.userData
    
                self.spawn_particles(point[0], point[1], self.particle_spawner_collision_hammer_hammer, entity_a.player.emitter)
                self.spawn_particles(point[0], point[1], self.particle_spawner_collision_hammer_hammer, entity_b.player.emitter)
            else:
                self.hammer_hit_contact = True
            
        if sound != None:
            self.game.play_sound(sound)
    
    def EndContact(self, contact):
        self.remove_freshly_thrown_in_from_ball(contact.fixtureA)
        self.remove_freshly_thrown_in_from_ball(contact.fixtureB)
    
    def PreSolve(self, contact, oldManifold):
        if (contact.fixtureA.userData == None) or (contact.fixtureB.userData == None):
            return
        
        type_a, entity_a = contact.fixtureA.userData
        type_b, entity_b = contact.fixtureB.userData
        entities = ((type_a, entity_a), (type_b, entity_b))
        
        player = None
        player_goal = None
        hammer = None
        ball = None
        
        for (t, e) in entities:
            if isinstance(e, Player):
                if t == constants.USER_DATA_ENTITY:
                    player = e
                else:
                    player_goal = e
            elif isinstance(e, Hammer):
                hammer = e
            elif isinstance(e, Ball):
                ball = e                
        
        derived_player = player
        if (player == None) and (hammer != None):
            derived_player = hammer.player
        
        if ball != None:
            if derived_player != None:
                if (player != None) and (ball.last_touched_player == player):
                    contact.enabled = False
                else:
                    ball.touched_by_player(derived_player)
                    contact.restitution = 0
                    
            elif player_goal != None:
                spawner = self.particle_spawner_goal_left
                emitter = self.ball.get_current_emitter()
                #emitter = self.player_left.emitter
                
                if (player_goal == self.player_left):
                    spawner = self.particle_spawner_goal_right
                    #emitter = self.player_right.emitter
                    
                self.spawn_particles_at_contact_position(contact, spawner, emitter)
                self.ball_hit_goal(ball, player_goal)
    
    def PostSolve(self, contact, impulse):
        if not self.hammer_hit_contact:
            return
        
        if (contact.fixtureA.userData == None) or (contact.fixtureB.userData == None):
            return
        
        _, entity_a = contact.fixtureA.userData
        _, entity_b = contact.fixtureB.userData
        
        if isinstance(entity_a, Hammer) and isinstance(entity_b, Hammer):
            self.hammer_hit_contact = False
            
            worldManifold = contact.worldManifold
            
            point = worldManifold.points[0]
            point = utils.world_to_screen(point[0], point[1])
            
            #normal = worldManifold.normal
            #normal = (normal[0] * constants.WORLD_TO_SCREEN, -normal[1] * constants.WORLD_TO_SCREEN)

            #distance = impulse.normalImpulses[0] * 0.2
            
            #point = (point[0] + normal[0] * distance, point[1] + normal[1] * distance)
            
            velocity_a = entity_a.body.linearVelocity
            velocity_b = entity_b.body.linearVelocity
            
            velocity_sum = velocity_a + velocity_b
            velocity_sum.x *= constants.WORLD_TO_SCREEN
            velocity_sum.y *= -constants.WORLD_TO_SCREEN
            #print(velocity_sum.length)
            
            distance = constants.HAMMER_HAMMER_COLLISON_PARTICLE_OFFSET_DISTANCE
            point = (point[0] + velocity_sum.x * distance, point[1] + velocity_sum.y * distance)

            self.spawn_particles(point[0], point[1], self.particle_spawner_collision_hammer_hammer, entity_a.player.emitter)
            self.spawn_particles(point[0], point[1], self.particle_spawner_collision_hammer_hammer, entity_b.player.emitter)
    
    def screen_shake(self, distance, duration):
        if not self.config.graphical_effects >= 1:
            return
        
        if distance == 0:
            return
        
        self.game.screen_shake(distance, duration)
    
    def remove_freshly_thrown_in_from_ball(self, fixture):
        if fixture.userData == None:
            return
        
        e = fixture.userData[1]
        
        if not isinstance(e, Ball):
            return
        
        e.freshly_thrown_in = False
    
    def spawn_particles_at_contact_position(self, contact, spawner, emitter):
        worldManifold = contact.worldManifold
        point = worldManifold.points[0]
        point = utils.world_to_screen(point[0], point[1])
        
        self.spawn_particles(point[0], point[1], spawner, emitter)
    
    def spawn_particles(self, x, y, spawner, emitter):
        spawner.spawn(emitter, x, y)
        
    def player_won(self, winning_player):
        if self.game_over:
            return
        
        self.game.play_sound(constants.SOUND_GAME_OVER)
        
        self.game_over = True
        
        self.entities.remove(self.score_text_left)
        self.entities.remove(self.score_text_center)
        self.entities.remove(self.score_text_right)
        self.overlay_entities.add(self.score_text_left)
        self.overlay_entities.add(self.score_text_center)
        self.overlay_entities.add(self.score_text_right)
        
        won_text = engine.Text(winning_player.name + " won!", self.game.font_big[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2, constants.WINDOW_HEIGHT / 2), winning_player.color,
                               engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_CENTER)
        self.overlay_entities.add(won_text)
        
        restart_string = "Press Space "
        back_string = "Press Escape "
        
        if len(self.game.joysticks) > 0:
            restart_string += "or A/Start "
            back_string += "or B/Back "
        
        restart_string += "to restart."
        back_string += "to go back to the menu."
        
        offset = 10
        y = constants.WINDOW_HEIGHT - offset
        color = (0, 0, 0)
        
        back_text = engine.Text(back_string, self.game.font_small[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2, y), color, engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_BOTTOM)
        self.overlay_entities.add(back_text)
        
        y = y - back_text.rect.height - offset
        restart_text = engine.Text(restart_string, self.game.font_small[self.config.skin], data.Point(constants.WINDOW_WIDTH / 2, y), color, engine.Text.H_ALIGN_CENTER, engine.Text.V_ALIGN_BOTTOM)
        self.overlay_entities.add(restart_text)
        
    
    def ball_hit_goal(self, ball, scoring_player):
        if self.ball_remove:
            return
        
        self.ball_remove = True
        
        own_goal = False
        
        if ball.last_touched_player == scoring_player:
            scoring_player.score += 2
        else:
            scoring_player.score += 1
            own_goal = True
            
        self.update_score()
        
        if scoring_player.score >= self.config.duration:
            self.player_won(scoring_player)
        else:
            if own_goal:
                self.game.play_sound(constants.SOUND_OWN_GOAL)
            else:
                self.game.play_sound(constants.SOUND_GOAL)
    
    def throw_ball(self):
        self.game.play_sound(constants.SOUND_BALL_THROWN_IN)
        
        start_y = 0
        start_velocity_y = -(constants.BALL_THROW_IN_MIN_SPEED + random.random() * (constants.BALL_THROW_IN_MAX_SPEED - constants.BALL_THROW_IN_MIN_SPEED)) * constants.SCREEN_TO_WORLD
        
        if random.random() < 0.5:
            start_y = constants.WINDOW_HEIGHT - start_y
            start_velocity_y *= -1
        
        trail_step_width = 1
        
        self.ball = Ball(self, data.Point(constants.WINDOW_WIDTH / 2, start_y), self.ball_default_emitter, self.ball_trail_default_emitter, trail_step_width)
        self.ball.body.ApplyLinearImpulse(Box2D.b2Vec2(0, start_velocity_y), self.ball.body.position)
        self.entities.add(self.ball)

    def update_score(self):
        self.score_text_left.text = str(self.players[0].score)
        self.score_text_right.text = str(self.players[1].score)

    def pause(self):
        if self.paused:
            return
        
        pygame.mixer.pause()
        
        self.paused = True
        
        self.game.play_sound(constants.SOUND_PAUSE)
        
        for pause_entity in self.pause_entities:
            self.overlay_entities.add(pause_entity)

    def resume(self):
        if not self.paused:
            return
        
        pygame.mixer.unpause()
        
        self.paused = False
        
        self.game.play_sound(constants.SOUND_RESUME)

        for pause_entity in self.pause_entities:
            self.overlay_entities.remove(pause_entity)

    def restart(self):
        self.game.state = PlayState(self.game, self.config)
        
    def back_to_menu(self):
        self.game.play_sound(constants.SOUND_MENU_ACTION)
        self.game.state = menu_state.MenuState(self.game, self.config)


# The main player entity, steerable by the user
class Player(engine.Entity):
    
    def __init__(self, play_state, name, color_name, color, filter_category, position, rotate, hammer_start_offset, hammer_rope_length_subtraction, hammer_max_distance_world, emitter, ball_trail_emitter,
                 key_left, key_right, key_up, key_down, joystick_id):
        engine.Entity.__init__(self)
        self.play_state = play_state
        self.game = play_state.game
        self.name = name
        self.color_name = color_name
        self.color = color
        self.filter_category = filter_category
        self.rotate = rotate
        self.hammer_rope_length_subtraction = hammer_rope_length_subtraction
        self.emitter = emitter
        self.ball_trail_emitter = ball_trail_emitter
        self.key_left = key_left
        self.key_right = key_right
        self.key_up = key_up
        self.key_down = key_down
        self.joystick_id = joystick_id
        self.direction = data.Point()
        
        self.set_image(utils.load_png('player_' + color_name + '.png', self.play_state.config.skin))
        self.set_position(position.x, position.y)
        #self.center()

        self.radius = (self.rect.width / 2)
        self.radius_world = self.radius * constants.SCREEN_TO_WORLD
        
        body_def = Box2D.b2BodyDef()
        body_def.position = utils.screen_to_world(position.x, position.y)
        body_def.type = Box2D.b2_dynamicBody
        #body_def.type = Box2D.b2_kinematicBody
        body_def.gravityScale = constants.PLAYER_GRAVITY_SCALE
        body_def.fixedRotation = True
        self.body = self.play_state.world.CreateBody(body_def)
        
        shape = Box2D.b2CircleShape(radius=self.radius_world)
        
        mask_bits = constants.FILTER_BORDER | constants.FILTER_BALL
        
        fixture = Box2D.b2FixtureDef(shape=shape,
                                     density=1,
                                     friction=constants.PLAYER_FRICTION,
                                     categoryBits=(constants.FILTER_PLAYER | self.filter_category),
                                     maskBits=mask_bits,
                                     userData=(constants.USER_DATA_ENTITY, self))
        
        self.body.CreateFixture(fixture)

        self.hammer = Hammer(self.play_state, self, position.add(hammer_start_offset))
        #self.play_state.entities.add(self.hammer)
        
        #if constants.DEBUG and color_name == "blue": self.hammer.fixture.filterData.maskBits = 0
        
        self.score = 0
        
        if hammer_max_distance_world > 0:
            rope_joint_def = Box2D.b2RopeJointDef()
            rope_joint_def.bodyA = self.body
            rope_joint_def.bodyB = self.hammer.body
            rope_joint_def.localAnchorA=(0,0) 
            rope_joint_def.localAnchorB=(0,0)
            rope_joint_def.maxLength = hammer_max_distance_world
            self.play_state.world.CreateJoint(rope_joint_def)


    def update(self):
        self.body.linearVelocity = ((self.direction.x * constants.PLAYER_FORCE, -self.direction.y * constants.PLAYER_FORCE))
        self.apply_body_to_entity(False)
        
        self.hammer.update()
        
        delta_to_hammer = self.hammer.position.subtract(self.position)
        angle_to_hammer = -math.degrees(delta_to_hammer.angle())
        if self.rotate:
            self.set_rotation(angle_to_hammer)
        self.hammer.set_rotation(angle_to_hammer + 180)
        
        initial_length = constants.PLAYER_HAMMER_INITIAL_DISTANCE_WORLD
        stiffness = constants.PLAYER_HAMMER_STIFFNESS
        
        position_player = self.body.position
        position_hammer = self.hammer.body.position
        delta_vector = position_hammer - position_player
        distance = delta_vector.length
        if distance > initial_length:
            delta_length = distance - initial_length
            force = stiffness * delta_length
            
            # Normalize or set
            if distance == 0:
                delta_vector = Box2D.b2Vec2(0.70710678118654757, 0.70710678118654757)
            else:
                delta_vector = Box2D.b2Vec2(delta_vector.x / distance, delta_vector.y / distance)
                
            spring_force = Box2D.b2Vec2(delta_vector.x * force, delta_vector.y * force)
            #self.body.ApplyForce(spring_force, position_player)
            self.hammer.body.ApplyForce(-spring_force, position_hammer)
            
    def draw(self, surface):
        engine.Entity.draw(self, surface)
        
        if self.hammer_rope_length_subtraction == 0:
            pygame.draw.line(surface, self.color, self.position.as_tuple(), self.hammer.position.as_tuple(), 3)
        else:
            rope_vector = self.hammer.position.subtract(self.position)
            rope_vector_length = rope_vector.length()
            
            if rope_vector_length > self.hammer_rope_length_subtraction:
                rope_vector.normalize()
                rope_vector.multiply(rope_vector_length - self.hammer_rope_length_subtraction)
                pygame.draw.line(surface, self.color, self.position.as_tuple(), (self.position.x + rope_vector.x, self.position.y + rope_vector.y), 3)
        
        self.hammer.draw(surface)

    def handle_event(self, event):
        
        if event.type == pygame.JOYAXISMOTION:
            if event.joy != self.joystick_id:
                return
            
            if event.axis % 2 == 0:
                self.direction.x = event.value
            elif event.axis % 2 == 1:
                self.direction.y = event.value
            
            if self.direction.length() < constants.JOYSTICK_DEAD_ZONE:
                self.direction.x = 0
                self.direction.y = 0
        
        elif event.type == pygame.JOYHATMOTION:
            if event.joy != self.joystick_id:
                return
            
            self.direction.x = event.value[0]
            self.direction.y = -event.value[1]
            self.direction.normalize()
        
        # Only react on key down and key up events
        elif (event.type == pygame.KEYDOWN) or (event.type == pygame.KEYUP):
            # Reset direction
            self.direction.x = 0
            self.direction.y = 0
            
            # Set direction according to which keys are currently down
            if self.key_left in self.game.keys_down:
                self.direction.x -= 1
                
            if self.key_right in self.game.keys_down:
                self.direction.x += 1

            if self.key_up in self.game.keys_down:
                self.direction.y -= 1

            if self.key_down in self.game.keys_down:
                self.direction.y += 1
            
            # Normalize the direction
            self.direction.normalize()
            
            self.body.linearVelocity = ((self.direction.x * constants.PLAYER_FORCE, -self.direction.y * constants.PLAYER_FORCE))
    

class Hammer(engine.Entity):
    
    def __init__(self, play_state, player, position):
        engine.Entity.__init__(self)
        self.play_state = play_state
        self.game = play_state.game
        self.player = player
        
        self.set_image(utils.load_png('hammer_' + player.color_name + '.png', self.play_state.config.skin))
        self.set_position(position.x, position.y)
        #self.center()

        self.radius = (self.rect.width / 2)
        self.radius_world = self.radius * constants.SCREEN_TO_WORLD

        body_def = Box2D.b2BodyDef()
        body_def.position = utils.screen_to_world(position.x, position.y)
        body_def.type = Box2D.b2_dynamicBody
        body_def.linearDamping = constants.HAMMER_LINEAR_DAMPING
        body_def.gravityScale = constants.HAMMER_GRAVITY_SCALE
        body_def.fixedRotation = True
        self.body = self.play_state.world.CreateBody(body_def)
        
        shape = Box2D.b2CircleShape(radius=self.radius_world)
        
        mask_bits = constants.FILTER_BALL
        if constants.HAMMERS_COLLIDE:
            mask_bits |= constants.FILTER_HAMMER
        if constants.HAMMERS_COLLIDE_WITH_BORDER:
            mask_bits |= constants.FILTER_BORDER
        
        fixtureDef = Box2D.b2FixtureDef(shape=shape,
                                     density=1,
                                     friction=constants.HAMMER_FRICTION,
                                     categoryBits=constants.FILTER_HAMMER,
                                     maskBits=mask_bits,
                                     userData=(constants.USER_DATA_ENTITY, self))
        
        self.fixture = self.body.CreateFixture(fixtureDef)


    def update(self):
        self.apply_body_to_entity(False)


# Just a ball flying around. Nothing more, nothing less.    
class Ball(engine.Entity):
    def __init__(self, play_state, position, default_emitter, ball_trail_emitter, trail_step_width):
        engine.Entity.__init__(self)
        self.play_state = play_state
        self.game = play_state.game
        self.velocity = data.Point()
        self.default_emitter = default_emitter
        self.ball_trail_emitter = ball_trail_emitter
        self.trail_spawner = particles.ParticleSpawnerTrail(constants.BALL_TRAIL_DURATION, constants.BALL_TRAIL_FADE_OUT_DURATION, trail_step_width, position.x, position.y)
        
        self.default_mask_bits = constants.FILTER_BORDER | constants.FILTER_HAMMER
         
        if constants.PLAYER_CAN_HIT_BALL:
            self.default_mask_bits |= constants.FILTER_PLAYER
        
        self.fixture = None
        
        self.freshly_thrown_in = True
        
        self.touched_by_player(None)
        self.set_position(position.x, position.y)
        #self.center()

        self.radius = (self.rect.width / 2)
        self.radius_world = self.radius * constants.SCREEN_TO_WORLD

        body_def = Box2D.b2BodyDef()
        body_def.position = utils.screen_to_world(position.x, position.y)
        body_def.type = Box2D.b2_dynamicBody
        body_def.gravityScale = constants.BALL_GRAVITY_SCALE
        body_def.linearDamping = constants.BALL_LINEAR_DAMPING
        body_def.bullet = True
        body_def.fixedRotation = True
        self.body = self.play_state.world.CreateBody(body_def)
        
        shape = Box2D.b2CircleShape(radius=self.radius_world)
        
        fixtureDef = Box2D.b2FixtureDef(shape=shape,
                                     density=constants.BALL_DENSITY,
                                     friction=constants.BALL_FRICTION,
                                     restitution=constants.BALL_RESTITUTION,
                                     categoryBits=constants.FILTER_BALL,
                                     maskBits=0,
                                     userData=(constants.USER_DATA_ENTITY, self))
        
        self.fixture = self.body.CreateFixture(fixtureDef)
        
        self.update_fixture()

    def touched_by_player(self, player):
        self.last_touched_player = player
        
        self.current_ball_trail_emitter = self.ball_trail_emitter
        
        color_name = "neutral"

        if player != None:
            color_name = player.color_name
            self.current_ball_trail_emitter = player.ball_trail_emitter
            
        self.set_image(utils.load_png('ball_' + color_name + '.png', self.play_state.config.skin))
        
        if self.fixture != None:
            self.update_fixture()
            
    def update_fixture(self):
        mask_bits = self.default_mask_bits

        if (constants.PLAYER_CAN_HIT_ENEMY_COLORED_BALL) and (self.last_touched_player != None):
            for player in self.play_state.players:
                if player != self.last_touched_player:
                    mask_bits |= player.filter_category
        
        self.fixture.filterData.maskBits = mask_bits

    def get_current_emitter(self):
        if (self.last_touched_player != None):
            return self.last_touched_player.emitter
        else:
            return self.default_emitter

    def update(self):
        #self.body.ApplyForceToCenter(Box2D.b2Vec2(0, 0.3))
        self.apply_body_to_entity(True)

        if self.current_ball_trail_emitter != None:
            self.trail_spawner.spawn(self.current_ball_trail_emitter, self.position.x, self.position.y)
